#include <stdio.h>
#include <tiuser.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netdb.h>
#include <stropts.h>
#include "tpinfo.h"

#define READ 1
#define WRITE 2

#define MAX_DATA  9999
#define HDR_SIZE  4

#define BSIZE MAX_DATA + HDR_SIZE

static int mode;             /* holds r or w indication   */
static int amount_in_buffer; /* amt in server buffer      */

int
rpopen(host, command, type)
char *host;
char *command;
char *type;
{
   int fd;              /* holds the transport endpoint   */
   int nbytes;          /* used to read user name         */
   int flags = 0;       /* info returned from t_rcv()     */
   int got_connection;  /* 1 if connection is established */
   int num_addrs;       /* num of address for a provider  */
   char buf[BUFSIZ];    /* generic buffer to hold data    */ 
   char prompt[BUFSIZ]; /* holds password prompt          */

   struct t_call *sndcall;      /* used in t_connect()    */
   struct tpinfo *tpinfop;      /* return from get_addrs()*/
   struct tpinfo *curr_tpinfop; /* traverses tpinfop      */

   if ((strcmp(type, "r") != 0)
    && (strcmp(type, "w") != 0)) {
      return(-1);
   }

   if (strcmp(type, "r") == 0) {
      mode = READ;
   } else {
      mode = WRITE;
   }

   /*
    *  Get the address and transport providers to use.
    */

   if ((tpinfop = get_addrs(host, "rpopen")) == NULL) {
      fprintf(stderr,
              "There are no addresses for this service!\n");
      return(-1);
   }

   got_connection = 0;
   curr_tpinfop = tpinfop;
   for (curr_tpinfop = tpinfop; curr_tpinfop;
                       curr_tpinfop = curr_tpinfop->nextp) {
      if ((fd = t_open(curr_tpinfop->device,
                                       O_RDWR, NULL)) < 0) {
         continue;
      }

      /*
       *  Bind to an arbitrary return address.
       */
   
      if (t_bind(fd, NULL, NULL) < 0) {
         (void)t_close(fd);
         continue;
      }
   
      /*
       *  Get memory to hold call structure, copy in
       *  the server's address, and connect to the service.
       */
   
      if ((sndcall = (struct t_call *)t_alloc(fd,
                                     T_CALL, 0)) == NULL) {
         (void)t_close(fd);
         continue;
      }
   
      /*
       *  Try to connect to the server by looping
       *  through all of the transport addresses.
       */
   
      num_addrs = curr_tpinfop->nd_addrlistp->n_cnt;
      while (num_addrs --) {
         sndcall->addr = *(tpinfop->nd_addrlistp->n_addrs);
   
         if (t_connect(fd, sndcall, NULL) == 0) {
            got_connection = 1;
            break;
         }
         curr_tpinfop->nd_addrlistp->n_addrs++;
      }
   
      if (got_connection) {
         break;
      }
      sndcall->addr.buf = 0;
      t_free(sndcall, T_CALL);
      (void)t_close(fd);
      curr_tpinfop = curr_tpinfop->nextp;
   }

   free_addrs(tpinfop);
   sndcall->addr.buf = 0;
   t_free(sndcall, T_CALL);

   if (!got_connection) {
      (void)t_close(fd);
      return(-1);
   }

   /*
    *  Read the user name, strip the newline at the
    *  end of the string, and send the name to the
    *  server side of the application.
    */

   printf("Please enter your user name of %s:\n", host);
   if ((nbytes = read(0, buf, BUFSIZ)) < 0) {
      (void)t_close(fd);
      return(-1);
   }

   buf[strlen(buf) - 1] = '\0';
   if (t_snd(fd, buf, strlen(buf) + 1, 0) == -1) {
      (void)t_close(fd);
      return(-1);
   }

   /*
    *  Prompt for the password and send it to the
    *  server side of the application.
    */

   sprintf(prompt,
               "Please enter your password on %s:\n", host);
   strcpy(buf, getpass(prompt));
   if (t_snd(fd, buf, strlen(buf) + 1, 0) == -1) {
      (void)t_close(fd);
      return(-1);
   }

   /*
    *  Read the results of the password validation.
    */

   if (t_rcv(fd, buf, 1, &flags) == -1) {
      (void)t_close(fd);
      return(-1);
   }
   if (buf[0] == 'n') {
      fprintf(stderr, "Login Incorrect\n");
      (void)t_snddis(fd, NULL);
      (void)t_close(fd);
      return(-1);
   }

   /*
    *  Send the type.
    */

   buf[0] = *type;
   if (t_snd(fd, buf, 1, 0) == -1) {
      (void)t_close(fd);
      return(-1);
   }

   /*
    *  Send the command.
    */

   if (t_snd(fd, command, strlen(command) + 1, 0) == -1) {
      (void)t_close(fd);
      return(-1);
   }

   return(fd);
}

int
rpread(fd, buf, len)
int fd;
char *buf;
int len;
{
   static char from_server[BSIZE];  /* holds server data */
   static int start_of_data = 0;    /* where data start  */

   int nbytes;          /* used to read user name         */
   int flags = 0;       /* info returned from t_rcv()     */
   int amount_to_get;   /* how much more data to read     */
   int user_start;      /* current offset in user buffer  */
   int where;           /* used to get data from server   */

   if (mode == WRITE) {
      return(-1);
   }

   user_start = 0;
   amount_to_get = len;
   while (amount_to_get) {
      /*
       *  The amount_in_buffer variable is a global
       *  variable initially set to 0.  If there are enough
       *  residual data from previous call to satisfy
       *  this request, fill the user buffer and break out
       *  of the loop.
       */
      if (amount_in_buffer >= amount_to_get) {
         memcpy(&buf[user_start],
                &from_server[start_of_data], amount_to_get);
         start_of_data += amount_to_get;
         amount_in_buffer -= amount_to_get;
         break;
      }

      /*
       *  Copy any residual data from previous call into the
       *  user buffer.
       */

      if (amount_in_buffer) {
         memcpy(&buf[user_start],
              &from_server[start_of_data], amount_in_buffer);
         amount_to_get -= amount_in_buffer;
         user_start += amount_in_buffer;
         amount_in_buffer = 0;
      }

      /*
       *  Get the first 4 bytes (which contain
       *  the length indicator).  If we fail (because the
       *  connection was closed), return the amount we
       *  already put into the buffer.
       */

      start_of_data = 0;
      where = 0;
      do {
         if ((nbytes = t_rcv(fd, &from_server[where],
                           HDR_SIZE - where, &flags)) < 0) {
            return(len - amount_to_get);
         }
         where += nbytes;
      } while (where < HDR_SIZE);

      (void)sscanf(from_server, "%4d", &amount_in_buffer);

      /*
       *   Check if this packet is the end-of-data marker.
       *   If it is, send the disconnect request to close the
       *   connection.
       */

      if (amount_in_buffer == 0) {
         (void)t_snddis(fd, NULL);
         (void)t_close(fd);
         return(len - amount_to_get);
      }

      /*
       *  Because we are not at the end-of-data marker,
       *  read the data in this packet and put them into
       *  our buffer.
       */

      where = 0;
      do {
         if ((nbytes = t_rcv(fd, &from_server[where],
                   amount_in_buffer - where, &flags)) < 0) {
            (void)t_close(fd);
            amount_in_buffer = where;
            break;
         }
         where += nbytes;
      } while (where < amount_in_buffer);
   }
   return(user_start + amount_to_get);
}

int
rpwrite(fd, buf, len)
int fd;
char *buf;
int len;
{
   char sendbuff[BSIZE]; /* buffer to send to server      */
   int amount;           /* amount of data to send        */
   int num_sent = 0;     /* amount of data sent           */

   if (mode == READ) {
      return(-1);
   }
   while (len) {
      amount = len;
      if (amount > MAX_DATA) {
         amount = MAX_DATA;
      }
      sprintf(sendbuff, "%.4d", amount);
      memcpy(sendbuff + HDR_SIZE, buf, amount);

      if (t_look(fd) == T_DISCONNECT) {
          fprintf(stderr, "disconnect occurred!\n");
          return(-1);
      }

      if (t_snd(fd, sendbuff, amount + HDR_SIZE, 0) < 0) {
         break;
      }
      len -= amount;
      num_sent += amount;
   }
   return(num_sent);
}

void
rpclose(fd)
int fd;
{
   int flags = 0;       /* info returned from t_rcv()     */

   if (mode == 0) {
      return;
   }
   if (mode == READ) {
      amount_in_buffer = 0;
   } else {
      /*
       *  Send the "0000" string to indicate the end of the
       *  data and wait for the server side to disconnect.
       */
      if (t_snd(fd, "0000", HDR_SIZE, 0) != HDR_SIZE) {
         t_error("t_snd of 0 failed!");
         mode = 0;
         (void)t_close(fd);
         return;
      }
      (void)t_rcv(fd, &mode, 1, &flags);
      exit(0);
   }

   mode = 0;
   (void)t_close(fd);
}
